# zettelkasten/monkeypatch.py
"""
Module aggregating all of the monkey patching functionalities.

Each time a command is executed via the command line,
zettelkasten monkey patches it's defaults as specified
in it's :ref:`cfile`.
"""

from typing import Any, MutableMapping, Optional, Type, Union

import configparser
import logging
from datetime import date

from . import defaults

logger = logging.getLogger(__name__)


def patch_defaults(config_file_path):
    """
    Main monkeypatching utility. Ensures the paramters inside the
    :ref:`cfile` are enforced by overwriting zettelkasten's defaults.
    """

    # create a config parse able to parse lists
    configs = configparser.ConfigParser(
        converters={"list": lambda x: [i.strip() for i in x.split(",")]}
    )

    # read in the config file
    configs.read(config_file_path)

    value: Optional[str] = None  # add typing hint for value
    for key, value in configs["default"].items():
        if key in defaults.config_overwrites:
            logger.debug(f"Monkeypatching {key}\n")
            logger.debug("with")
            logger.debug(f"{value}\n")
            if value == "None":
                value = None

            setattr(defaults, key, value)

    # parse pure list dict
    source_file_formats = {}
    for key in configs["source_file_formats"]:
        source_file_formats[key] = configs["source_file_formats"].getlist(key)

    logger.debug("Monkeypatching source_file_formats\n")
    logger.debug("with")
    logger.debug(f"{source_file_formats}\n")
    defaults.source_file_formats = source_file_formats

    # parse pure string dict
    zettel_meta_attribute_labels = {
        k: v for k, v in configs["zettel_meta_attribute_labels"].items()
    }

    logger.debug("Monkeypatching zettel_meta_attribute_labels\n")
    logger.debug("with")
    logger.debug(f"{zettel_meta_attribute_labels}\n")
    defaults.zettel_meta_attribute_labels = zettel_meta_attribute_labels

    # pare mixed dict
    zettel_meta_attribute_defaults: MutableMapping[str, Any] = {}
    values = ["category", "subcategory", "author", "doc", "dole"]
    for key in configs["zettel_meta_attribute_defaults"]:
        if key in values:
            v: Optional[str] = configs["zettel_meta_attribute_defaults"][key]
            if v == "None":
                v = None
                zettel_meta_attribute_defaults[key] = v
            if v in ["today", "now", "jetzt", "heute"]:

                date.today().strftime("%Y-%m-%d")
                d: Type[date] = date
                zettel_meta_attribute_defaults[key] = d

        else:
            lst = configs["zettel_meta_attribute_defaults"].getlist(key)
            if lst == [""]:
                lst = list()
            zettel_meta_attribute_defaults[key] = lst

    logger.debug("Monkeypatching zettel_meta_attribute_defaults\n")
    logger.debug("with")
    logger.debug(f"{zettel_meta_attribute_defaults}\n")
    defaults.zettel_meta_attribute_defaults = zettel_meta_attribute_defaults

    logger.debug("Set the defaults.state 'config_file_monkeypatched'")
    defaults.state = "config_file_monkeypatched"



# type: ignore[attr-defined]
"""CLI interface for managing a folder based zettelkasten using emacs org-mode
file zettels and a bibtex reference system"""

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



# zettelkasten/add.py
import typing

import logging
import os
import pathlib
import shutil

from . import defaults

logger = logging.getLogger(__name__)


class ZettelName(typing.NamedTuple):
    """
    Zettel name consisting of ``category``, ``subcategory``, ``uid``

    Parameters
    ----------
    category: str
        String representing the top level folder inside the zettelkasten.
       ``None`` if Zettel is inside the :ref:`Lobby`.

    subcategory: str
        String representing the top sublevel folder inside the zettelkasten.
        ``None`` if no subcategory is given.

    uid: str
        String representing the top sublevel folder inside the zettelkasten.
        ``None`` if no subcategory is given.

    Return
    ------
    zettel_name: ~typing.NamedTuple
        Namedtuple instance object serving as zettel name.

    Examples
    --------
    >>> ZettelName('folder', 'subfolder', 'zettel')
    ZettelName(category='folder', subcategory='subfolder', uid='zettel')
    """

    category: str
    subcategory: str
    uid: str


def parse_zettel_name(name):
    r"""Zettel name parsing syntax

    Parameters
    ----------
    name : str
        Zettel name. Either a singular uid as in "my_zettel" or a
        folder/subfolder/uid syntax as in "woodturning/tools/chisel".

        Used to automatically sort new zettels either into the
        category/subcategory stated or into the :ref:`Lobby` if only an uid is
        given.

        The seperator used for uidentifying the different parts is specified
        using the :ref:`cfile`. It defaults to ``\``.

        If no complete name is given, then category and subcategory are set to
        their defaults as stated in the :ref:`cfile`.


    Examples
    --------
    1. Using a complete name to automatically sort the zettel:

        >>> parse_zettel_name("woodturning/tools/chisel")
        ZettelName(category='woodturning', subcategory='tools', uid='chisel')

    2. Using only an uid, puts the Zettel into the :ref:`Lobby`:

        >>> parse_zettel_name("my_zettel")
        ZettelName(category=None, subcategory=None, uid='my_zettel')

    3. Using the wrong syntax fails somewhat gracefully:

        >>> parse_zettel_name("wrong_syntax/°_^%$!§")
        ZettelName(category=None, subcategory=None, uid='°_^%$!§')

    Returns
    -------
    zettel_name: tuple
        Tuple consisting of ('category', 'subcategory', 'uid')
    """

    logger.debug("-------------------------------")
    logger.debug("Preparing to parse Zettel name")

    # is a complete name given?
    if defaults.name_sep in name:

        # yes, to parse it
        logger.debug(f"{defaults.name_sep} found in '{name}'")
        parts = name.split(defaults.name_sep)

        # syntax is only valid using 2 seperators:
        if len(parts) == 3:
            category, subcategory, uid = parts
            logger.debug(f"Succesfully dissassembled '{name}' into:")
            logger.debug(
                f"category: '{category}', subcategory: "
                + f"'{subcategory}', uid: '{uid}'"
            )

        # used wrong syntax. State a warning and fail gracefully
        else:
            uid = parts[-1]
            category = defaults.zettel_meta_attribute_defaults.get(
                "category", None
            )
            subcategory = defaults.zettel_meta_attribute_defaults.get(
                "category", None
            )

            logger.warning(f"'{name}' could not be dissassembled correctly")
            logger.warning(f"Dissassembling yielded '{parts}'. Parsing into:")
            logger.warning(
                f"category: '{category}', subcategory: "
                + f"'{subcategory}', uid: '{uid}'"
            )

    # No, only an uid was passed:
    else:
        uid = name
        category = defaults.zettel_meta_attribute_defaults.get("category", None)
        subcategory = defaults.zettel_meta_attribute_defaults.get(
            "category", None
        )

        logger.debug("Zettel name parsed.")
        logger.debug("-------------------------------\n")

    return ZettelName(category, subcategory, uid)


def parse_zettel_attributes(parsed_zettel_name, **kwargs):
    r"""
    Zettel attribute parsing utility.

    The attributes parsed are stated in the :ref:`cfile` along with a set
    of required attributes. These required attributes are made of the
    attributes of the :class:`ZettelName` and stem form this zettelkasten's
    dogma of having a ``zettelkasten/category/subcategory/zettel`` hierarchy.

    note
    ----
    Called by :meth:add_zettel during zettel creation.

    Parameters
    ----------
    parsed_zettel_name: ZettelName
        :class:`typing.NamedTuple` representing the zettel.

    **kwargs:
        The attributes to be parsed. Attributes are/need to be stated in the
        :ref:`cfile` along with a set of required attributes.

        Attributes will be added to the Zettel-Org-File as::

            # +Attribute_Name1: Attribute_Value1
                        .
                        .
            # +Attribute_NameN: Attribute_ValueN

    Return
    ------
    parsed_attributes: dict
        Dictionairy of the parsed attributes. The attributes parsed are stated
        in the :ref:`cfile`.

    Examples
    --------
    Building on the examples stated in :meth:`parse_zettel_name`

    >>> import pprint #imported for better doctest output

        1. Using a complete name to automatically sort the zettel:

        >>> zn = parse_zettel_name("woodturning/tools/chisel")
        >>> zattr = parse_zettel_attributes(zn)
        >>> pprint.pprint(zattr)
        {'#+Author:': 'Mathias Ammon',
         '#+Category:': 'woodturning',
         '#+DOC:': None,
         '#+DOLE:': None,
         '#+Subcategory:': 'tools',
         '#+Tags:': [],
         '#+Title:': 'chisel',
         '#+Topics:': []}


    2. Using only a uid, but adding tags and topics:

        >>> zn = parse_zettel_name("my_zettel")
        >>> zattr = parse_zettel_attributes(
        ...     zn,
        ...     tags=['#Rework', '#Pressurloss', '#Solarthermal'],
        ...     topics=['#EnergyConcept', '#Thermodynamics'])
        >>> pprint.pprint(zattr)
        {'#+Author:': 'Mathias Ammon',
         '#+Category:': None,
         '#+DOC:': None,
         '#+DOLE:': None,
         '#+Subcategory:': None,
         '#+Tags:': ['#Rework', '#Pressurloss', '#Solarthermal'],
         '#+Title:': 'my_zettel',
         '#+Topics:': ['#EnergyConcept', '#Thermodynamics']}


    3. Using the wrong syntax and using non supported attributes (unsupported):

        >>> zn = parse_zettel_name("wrong_syntax/°_^%$!§")
        >>> zattr = parse_zettel_attributes(
        ...     zn,
        ...     unsupported='This does not show')
        >>> pprint.pprint(zattr)
        {'#+Author:': 'Mathias Ammon',
         '#+Category:': None,
         '#+DOC:': None,
         '#+DOLE:': None,
         '#+Subcategory:': None,
         '#+Tags:': [],
         '#+Title:': '°_^%$!§',
         '#+Topics:': []}

    """

    logger.debug("-------------------------------")
    logger.debug("Preparing to parse zettel attributes")

    # map zettel atrributes
    zettel_attributes = dict()

    # start with the required ones using the zettel name:
    for req_at in defaults.required_attributes:

        zettel_attributes[
            defaults.zettel_meta_attribute_labels[req_at]
        ] = getattr(parsed_zettel_name, req_at)

    # add to that the optional ones as stated in the cfile
    for attribute, label in defaults.zettel_meta_attribute_labels.items():

        if attribute not in defaults.required_attributes:
            zettel_attributes[label] = kwargs.get(
                attribute, defaults.zettel_meta_attribute_defaults[attribute]
            )

    return zettel_attributes


def create_zettel_location(parsed_zettel_name, dummy_location=None):
    """
    Utility for making sure the requested zettel location exists.

    The zettelkasten location is stated in the :ref:`cfile`.

    In case the :paramref:`~create_zettel_location.parsed_zettel_name` states
    a :paramref:`~ZettelName.category` this serves as toplevel folder
    inside the zettelkasten folder and the :paramref:`~ZettelName.subcategory`
    serves as folder inside the toplevel folder.

    The Zettel itself is then represented as a folder inside this subcategory
    folder and storing all the necessary files for taking advantage of the
    org file.

    In case no :paramref:`~ZettelName.category` is stated (and therfor also no
    :paramref:`~ZettelName.subcategory`), the zettel folder is created inside
    the :ref:`lobby` which is a folder directly beneath the zettelkasten folder
    of which it's name is specified again inside the :ref:`cfile`. The default
    lobby folder name is ``lobby``.

    note
    ----
    Called by :meth:add_zettel during zettel creation.

    Parameters
    ----------
    parsed_zettel_name: ZettelName
        :class:`NamedTuple` representing the zettel location and uid as
        returned by :meth:`parse_zettel_name`

    dummy_location: str, pathlib.Path, None, default=None
        Dummy location used for testing.

    Return
    ------
    zettel_location: pathlib.Path
        Path specifying the zettel location.

    Examples
    --------
    Building on the examples stated in :meth:`parse_zettel_name`

    >>> import pprint #imported for better doctest output

        1. Using a complete name to automatically sort the zettel:

        >>> zn = parse_zettel_name("woodturning/tools/chisel")
        >>> path = create_zettel_location(
        ...     parsed_zettel_name=zn,
        ...     dummy_location="tests/testkasten/")
        >>> print(str(path))
        tests/testkasten/woodturning/tools/chisel

    """
    logger.debug(79 * "-")
    logger.debug("Preparing to create zettel location")
    folder = parsed_zettel_name.category
    subfolder = parsed_zettel_name.subcategory
    zettel_folder = parsed_zettel_name.uid

    if dummy_location is not None:
        zk_location = dummy_location
    else:
        zk_location = defaults.location
    # create zettel directory
    if folder is None and subfolder is None:
        logger.debug("Only zettel uid was stated.")
        logger.debug("Creating a lobby location for it")

        zettel_path = os.path.join(zk_location, "lobby", zettel_folder)

    else:
        logger.debug("Complete zettel name was stated")

        zettel_path = os.path.join(
            zk_location, folder, subfolder, zettel_folder
        )

    # create zettel path
    pathlib.Path(zettel_path).mkdir(parents=True, exist_ok=True)

    return pathlib.Path(zettel_path)


def write_org_zettel_attributes(org_file_path, zettel_attributes):
    r"""
    Utility wrapping the attribute writing in a singular call.

    Parameters
    ----------
    org_file_path: str, pathlib.Path
        Text or byte string stating the org zettel's file location.
        For more info on zettel location see :func:`create_zettel_location`.
    zettel_attributes: dict
        Dictionairy stating mapping attribute names to attribute values
        as stated in the :ref:`cfile`. For more on zettel attributes
        see :func:`parse_zettel_attributes`.

    Examples
    --------

    1. Create zettel folder dummy location:

       >>> import pathlib
       >>> pathlib.Path("tests/testkasten/lobby/attributes").mkdir(
       ...     parents=True, exist_ok=True)

    2. Create a dummy org file:

        >>> dummy_loc = "tests/testkasten/lobby/attributes/attributes.org"
        >>> write_org_zettel_attributes(
        ...     org_file_path=dummy_loc,
        ...     zettel_attributes={
        ...         'test1': 'test_attribute',
        ...         'test2': 'test_attribute',
        ...     }
        ... )

    3. Read out it's attributes:

        >>> with open(dummy_loc, 'r') as f:
        ...     pass
        ...     lines = f.readlines()
        ...     for line in lines:
        ...         print(line.strip(' \n'))
        test1 test_attribute
        test2 test_attribute
    """
    with open(org_file_path, "w") as f:
        for attribute, value in zettel_attributes.items():
            f.write(" ".join((attribute, str(value), "\n")))

    logger.debug(f"Wrote zettel_attributes: {zettel_attributes}")
    logger.debug(f"Into Zettel at: {org_file_path}")


def write_org_zettel_bibliography(
    org_file_path, bibliography_file, force_overwrite=False
):
    r"""
    Utility wrapping bibliography addition to the org file.

    Parameters
    ----------
    org_file_path: str, pathlib.Path
        Text or byte string stating the org zettel's file location.
        For more info on zettel location see :func:`create_zettel_location`.
    bibliography_file: str
        String specifying the bibliography file filename. File will be on the
        same level as :paramref:`~write_org_zettel_bibliography.org_file_path`
    force_overwrite: bool, default=False
        If ``True`` an already existing Zettel-Org-File will be overridden.
        Use with care.

    Examples
    --------

    1. Create zettel folder dummy location:

       >>> import pathlib
       >>> pathlib.Path("tests/testkasten/lobby/bibliography").mkdir(
       ...     parents=True, exist_ok=True)

    2. Create a dummy org file:

        >>> dummy_loc = (
        ...     "tests/testkasten/lobby/bibliography/bibliography.org"
        ... )
        >>> write_org_zettel_bibliography(
        ...     org_file_path=dummy_loc,
        ...     bibliography_file="bibliography.bib",
        ...     force_overwrite=True,
        ... )

    3. Read out it's bibliography content

        >>> with open(dummy_loc, 'r') as f:
        ...     lines = f.readlines()
        ...     for line in lines:
        ...         print(line.strip(' \n'))
        * Bibliography
        <BLANKLINE>
        bibliography:bibliography.bib

    """
    logger.debug("Writing bibliography section into zettel.")
    logger.debug("Assumed bibliography file name is")
    logger.debug(f"{bibliography_file}")

    if force_overwrite:
        mode = "w"
    else:
        mode = "a"

    with open(org_file_path, mode) as f:
        f.writelines(
            ["\n", "* Bibliography\n\n", f"bibliography:{bibliography_file}"]
        )


def create_bibliography_file(bibliography_file_path, force_overwrite=False):
    """
    Wrap bibliography file creation.

    Parameters
    ----------
    bibliography_file_path: str, pathlib.Path
        Text or byte string stating the bibliography's file location.
        For more info on the location see :func:`create_zettel_location`.
        By convention this has the same name as the corresponding
        zettel-org-file but with a different file ending.

    force_overwrite: bool, default=False
        If ``True`` an already existing Zettel-Bibliograph-File will be
        overridden. Use with care.

    Raises
    ------
    FileExistsError
        Raised if bibliography file already present and
        :paramref:`~create_bibliography_file.force_overwrite` is ``False``.
        Prevents unwanted data loss.

    Examples
    --------

    >>> import os

    1. Create zettel folder dummy location:

       >>> import pathlib
       >>> pathlib.Path("tests/testkasten/lobby/bibliography").mkdir(
       ...     parents=True, exist_ok=True)

    2. Create a dummy bib file:

        >>> dummy_loc = (
        ...     "tests/testkasten/lobby/bibliography/bibliography.bib"
        ... )
        >>> create_bibliography_file(
        ...     bibliography_file_path=dummy_loc,
        ...     force_overwrite=True)
        >>> os.path.isfile(dummy_loc)
        True
    """
    logger.debug(f"Writing bibliography file in {bibliography_file_path}.")

    # check for overwrite
    if os.path.isfile(bibliography_file_path):
        already_exists_msg = (
            f"Bibliography file in '{bibliography_file_path}' already exists"
        )

        if force_overwrite:
            logger.debug(already_exists_msg)
            logger.debug("Forced overwrite requested")
        else:  # dont force overwrite unless explicitly requested
            logger.error(already_exists_msg)
            logger.error("To purposely overwrite bibliography file use:")
            logger.error("zk add [-f/--force] zettel")
            raise FileExistsError

    with open(bibliography_file_path, "w"):
        pass


def create_bibliography_file_test_entries(
    zettel_name, force_overwrite=False, dummy_location=None
):
    """
    Wrap creating bibliography test entries.

    Parameters
    ----------
    zettel_name: str
        String specifying the name and/or the location of the zettel.
        See :class:`ZettelName` and :paramref:`parse_zettel_name.name` for more
        details.

    force_overwrite: bool, default=False
        If ``True`` already existing bibliography entries will be
        overridden. Use with care.

    dummy_location: str, pathlib.Path, None, default=None
        Dummy location used for testing. Design usage is to fallback on
        :attr:`zettelkasten.defaults.location`.
    """
    example_entries = [
        (
            "tests/bib_sources/test_audio.mp3",
            "audio_2021_sec2",
            "sec 2",
            "zk_api",
            "Test Audio",
        ),
        (
            "tests/bib_sources/test_image.jpg",
            "image_2021",
            None,
            "zk_api",
            "Test Image",
        ),
        (
            "tests/bib_sources/test_pdf.pdf",
            "pdf_2021_p2",
            "page 2",
            "zk_api",
            "Test PDF",
        ),
        (
            "tests/bib_sources/test_video.mp4",
            "video_2021_min42",
            "min 42",
            "zk_api",
            "Test Video",
        ),
    ]

    for (
        f,
        uid,
        locspec,
        author,
        title,
    ) in example_entries:
        new_source(
            zettel_name,
            f,
            uid,
            locspec=locspec,
            author=author,
            title=title,
            force_overwrite=force_overwrite,
            dummy_location=dummy_location,
        )


def new_zettel(name, force_overwrite=False, dummy_location=None, **kwargs):
    r"""Add a new Zettel to the Zettelkasten.

    Wraps all the utilities of:

        1. :func:`parse_zettel_name`
        2. :func:`parse_zettel_attributes`
        3. :func:`create_zettel_location`
        4. :func:`write_org_zettel_attributes`
        5. :func:`write_org_zettel_bibliography`

    in one callable designed to be used by the command line interface (cli).

    Parameters
    ----------
    name: str
        String specifying the name and/or the location of the zettel.
        See :class:`ZettelName` and :paramref:`parse_zettel_name.name` for more
        details.

    force_overwrite: bool, optional
        If ``True`` an already existing Zettel-Org-File will be overridden.
        Use with care. Default=False

    dummy_location: str, pathlib.Path, None,
        Dummy location used for testing. Used instead of
        :attr:`zettelkasten.defaults.location` when used.

    **kwargs
        Zettel attributes to be parsed. See :func:`parse_zettel_attributes`
        for more details.

    Raises
    ------
    FileExistsError
        Raised if zettel already present and
        :paramref:`~new_zettel.force_overwrite` is ``False``.

    Examples
    --------
    Building on the examples stated in :meth:`parse_zettel_name`

    >>> import os

        1. Using a complete name to automatically sort the zettel:

            >>> zn = parse_zettel_name("woodturning/tools/chisel")
            >>> path = create_zettel_location(
            ...     parsed_zettel_name=zn,
            ...     dummy_location="tests/testkasten/")
            >>> print(str(path))
            tests/testkasten/woodturning/tools/chisel

            (forced overwrite is used here for doctesting. Usually you don't
            want to do that)

            >>> new_zettel("woodturning/tools/chisel", force_overwrite=True)
            >>> os.path.isfile(
            ...     "tests/testkasten/woodturning/tools/chisel/chisel.org")
            True

        2. Use only a Zettel uid puts it into the lobby (stated in the
           :ref:`cfile`):

            (Again, forced overwrite is used here for doctesting. Usually you
            don't want to do that)

            >>> new_zettel("my_zettel", force_overwrite=True)
            >>> os.path.isfile(
            ...     "tests/testkasten/lobby/my_zettel/my_zettel.org")
            True

    Buildiing on the examples stated in :meth:`parse_zettel_attributes` tools
    add attributes to the Zettel during creation:

        >>> new_zettel(
        ...     name="woodturning/tools/gouge",
        ...     force_overwrite=True,
        ...     tags=["#Rework", "#NiceTry"],
        ...     topics=["#Test"],
        ...     not_showing="NotShowing",
        ...     )
        >>> zettel_org_file = pathlib.Path(
        ...     "tests/testkasten/woodturning/tools/gouge/gouge.org")
        >>> with open(zettel_org_file, 'r') as f:
        ...     lines = f.readlines()
        ...     for line in lines:
        ...         print(line.strip(' \n'))
        #+Title: gouge
        #+Category: woodturning
        #+Subcategory: tools
        #+Author: Mathias Ammon
        #+DOC: None
        #+DOLE: None
        #+Topics: ['#Test']
        #+Tags: ['#Rework', '#NiceTry']
        <BLANKLINE>
        * Bibliography
        <BLANKLINE>
        bibliography:gouge.bib
    """
    logger.debug("-------------------------------")
    logger.debug(f"Preparing to add Zettel '{name}'")

    # dissassemble the name/location syntax:
    zettel_name = parse_zettel_name(name)

    # parse the zettel attributes
    zettel_attributes = parse_zettel_attributes(zettel_name, **kwargs)

    zettel_path = create_zettel_location(
        parsed_zettel_name=zettel_name, dummy_location=dummy_location
    )

    # create zettel org file
    org_file_path = pathlib.Path(
        os.path.join(zettel_path, ".".join([zettel_name.uid, "org"]))
    )

    # check for overwrite
    if os.path.isfile(org_file_path):
        already_exists_msg = f"Zettel in '{zettel_path}' already exists"
        if force_overwrite:
            logger.debug(already_exists_msg)
            logger.debug("Forced overwrite requested")
        else:  # dont force overwrite unless explicitly requested
            logger.error(already_exists_msg)
            logger.error("To purposely overwrite a Zettel use:")
            logger.error("zk add [-f/--force] zettel")
            raise FileExistsError

    # wirte the zettel's inital attribute structure
    write_org_zettel_attributes(org_file_path, zettel_attributes)

    # create the zettel's bibliography file
    bib_file_path = pathlib.Path(
        os.path.join(zettel_path, ".".join([zettel_name.uid, "bib"]))
    )
    create_bibliography_file(bib_file_path, force_overwrite=force_overwrite)

    create_bibliography_file_test_entries(
        zettel_name=name,
        force_overwrite=True,
        dummy_location=dummy_location,
    )

    # write the zettel's bibiliography
    write_org_zettel_bibliography(org_file_path, f"{zettel_name.uid}.bib")

    logger.info(f"Succesfully created org-Zettel in '{org_file_path}'")


def write_source_entry(
    bibliography_file_path,
    source_file,
    uid,
    locspec,
    author,
    title,
    year,
    date,
    force_overwrite=False,
):
    """
    Utility wrapping source entry writing.

    Tests if the entry is already present,  whether to overwrite it or not and
    performs the actual writing.

    Parameters
    ----------
    bibliography_file_path: str, pathlib.Path
        Text or byte string stating the bibliography's file location.
        For more info on the location see :func:`create_zettel_location`.
        By convention this has the same name as the corresponding
        zettel-org-file but with a different file ending.

    source_file: file
        File representing the source that serves as reference.
        Supported file types are:

            - audio
            - image
            - pdf
            - video

        The source file is copied into :attr:`Zettelkasten's source folder
        <zettelkasten.defaults.sources_directory>`

    uid: str
        String uniquely specifying the source. Usually something like
        `author_year_locspec`

    locspec: str
        String specifying where (inside the source) exactly the reference
        points to. Something like:

            - 'page 42'
            - 'min 69',

    author: str
        String specifying the author. Either ``Surname, Name`` or
        ``Name Surname``

    title: str
        String specifying the source files title

    year: str, ~numbers.Number
        String or number specifying the year, the source was published

    date: str
        String specifying the date as in ``"%Y-%m-%d"``.

    force_overwrite: bool, default=False
        If ``True`` and the source is already present inside the bibliography
        file, then the entry will be overriden. Use with care.

    Raises
    ------
    FileExistsError
        Raised if source key already present and
        :paramref:`~write_source_entry.force_overwrite` is ``False``.
        Prevents unwanted data loss.
    """
    # open the zettel's bib file to write:
    logger.debug("Preparing to write the source entry")

    overwrite = False
    with open(bibliography_file_path) as f:
        content = f.read()
        if uid in content:
            already_exists_msg = (
                f"Entry of key {uid} already present in "
                + f"{bibliography_file_path}"
            )

            if force_overwrite:

                overwrite = True

                logger.debug(already_exists_msg)
                logger.debug("Overwrite requested.")

                part_1 = f"@misc{{{uid},\n"
                part_2 = content.split(part_1)[-1].split("\n}%\n")[0] + "\n}%\n"

                content_to_overwrite = part_1 + part_2

                replacement = "".join(
                    defaults.bibliography_entry(
                        source_file=source_file,
                        key=uid,
                        location_specifier=locspec,
                        author=author,
                        title=title,
                        year=year,
                        date=date,
                    )
                )

                content = content.replace(content_to_overwrite, replacement)
                # TODO overwrite
            else:
                logger.error(already_exists_msg)
                logger.error("To purposely overwrite an entry use:")
                # TODO Distinguish between api and command line call
                logger.error("zk source [-f/--force] file, key, [locspec]")
                raise FileExistsError

    if overwrite:
        with open(bibliography_file_path, "w") as f:
            logger.debug("Overwriting:\n")
            logger.debug(f"{content_to_overwrite}\n")
            logger.debug(f"in {bibliography_file_path}\n")
            logger.debug(f"with:\n {replacement}\n")
            f.write(content)

    else:
        with open(bibliography_file_path, "a") as f:
            logger.debug("No overwrite necessary, creating the entry")
            # print(f'writing into {bibliography_file_path}')
            f.writelines(
                defaults.bibliography_entry(
                    source_file=source_file,
                    key=uid,
                    location_specifier=locspec,
                    author=author,
                    title=title,
                    year=year,
                    date=date,
                )
            )


def new_source(
    zettel_name,
    source_file,
    uid,
    locspec=defaults.def_location_specifier,
    author=defaults.def_author,
    title=defaults.def_title,
    year=defaults.def_year,
    date=defaults.def_date,
    force_overwrite=False,
    dummy_location=None,
):
    r"""
    Source adding utility.

    Adds a source to a zettel's bibliography file as well as the
    :attr:`Zettelkasten's overall bibliography file
    <zettelkasten.defaults.zettelkasten_bib_file>`.
    Also copies the source file
    into the :attr:`Zettelkasten's source folder
    <zettelkasten.defaults.sources_directory>`.

    The bibliography entry is wirtten using the
    :attr:`bibliography entry <zettelkasten.defaults.bibliography_entry>`
    template.


    Parameters
    ----------
    zettel_name: str
        Zettel name. Either a singular uid as in "my_zettel" or a
        folder/subfolder/uid syntax as in "woodturning/tools/chisel". See
        :class:`ZettelName` and :func:`parse_zettel_name` for more details

    source_file: file
        File representing the source that serves as reference.
        Supported file types are:

            - audio
            - image
            - pdf
            - video

        The source file is copied into :attr:`Zettelkasten's source folder
        <zettelkasten.defaults.sources_directory>`

    uid: str
        String uniquely specifying the source. Usually something like
        `author_year_locspec`

    locspec: str
        String specifying where (inside the source) exactly the reference
        points to. Something like:

            - 'page 42'
            - 'min 69',

    author: str
        String specifying the author. Either ``Surname, Name`` or
        ``Name Surname``

    title: str
        String specifying the source files title

    year: str, ~numbers.Number
        String or number specifying the year, the source was published

    date: str
        String specifying the date as in ``"%Y-%m-%d"``.

    force_overwrite: bool, optional
        If ``True`` and the source is already present inside the bibliography
        file, then the entry will be overriden. Use with care.

    dummy_location: str, pathlib.Path, None, optional
        Dummy location used for testing. Design usage is to fallback on
        :attr:`zettelkasten.defaults.location`.

    Raises
    ------
    FileNotFoundError
        Raised when requested :paramref:`zettel <new_source.zettel_name>` is
        not found.

    Examples
    --------
    1. Default usecase, adding a new source (again, the force_overwrite=True
       is only used for doctesting, usually you don't want to use it this way):

        >>> new_zettel("woodturning/tools/skew", force_overwrite=True)
        >>> new_source(
        ...     zettel_name="woodturning/tools/skew",
        ...     source_file="tests/bib_sources/test_video.mp4",
        ...     uid="video2_2021_min42",
        ...     locspec="min 42",
        ...     force_overwrite=True)

        Check if the new source uid is in the zettels bibliography file:

        >>> with open(
        ...     "tests/testkasten/woodturning/tools/skew/skew.bib",
        ...     "r") as f:
        ...     "video2_2021_min42" in f.read()
        True

        Check if the file is now present inside the
        :attr:`Zettelkasten's source folder
        <zettelkasten.defaults.sources_directory>`:

        >>> os.path.isfile(
        ...     "tests/testkasten/_sources/videos/test_video.mp4")
        True

        See if the source uid is also present inside the zettelkasten's
        main bib file:

        >>> with open(
        ...     "tests/testkasten/_sources/zettelkasten.bib",
        ...     "r") as f:
        ...     "video2_2021_min42" in f.read()
        True
    """
    # generating the bib file string
    logger.debug("Preparing to add a new source entry")

    # dissassemble the name/location syntax:
    parsed_zettel_name = parse_zettel_name(zettel_name)

    # parse the zettel location
    zettel_path = create_zettel_location(
        parsed_zettel_name=parsed_zettel_name, dummy_location=dummy_location
    )

    org_file_path = pathlib.Path(
        os.path.join(zettel_path, ".".join([parsed_zettel_name.uid, "org"]))
    )

    # is zettel existing
    if not org_file_path.is_file():
        logger.error(f"Requested zettel was not found in {org_file_path}")
        raise FileNotFoundError

    # yes zettel is existing, so infer the zettel bibliography file path
    bib_file_path = pathlib.Path(
        os.path.join(zettel_path, ".".join([parsed_zettel_name.uid, "bib"]))
    )

    if dummy_location:
        location = dummy_location
    else:
        location = defaults.location

    logger.debug("Copying the new source entry into to main bib file")
    # also write an entry into the zettelkasten's bib file:
    zk_bib_file = os.path.join(
        location,
        defaults.sources_directory,
        defaults.zettelkasten_bib_file,
    )

    # copying the source file
    # find out directory by inspecting fileendings
    ftype = defaults.infer_file_type(source_file)

    destination = os.path.join(
        location,
        # defaults.location,
        defaults.sources_directory,
        ftype,
        source_file.split("/")[-1],
    )

    logger.debug(f"Copying the source file into {destination}")

    # and copy the file including permissions and meta data
    shutil.copy2(
        source_file,
        destination,
    )

    logger.debug("Writing the entry into the main bib file:")
    # also write the entry into the zk bib file
    write_source_entry(
        zk_bib_file,
        source_file=destination,
        uid=uid,
        force_overwrite=force_overwrite,
        locspec=locspec,
        author=author,
        title=title,
        year=year,
        date=date,
    )

    # write the actual source entry
    logger.debug("Writing the actual entry:")
    write_source_entry(
        bibliography_file_path=bib_file_path,
        source_file=destination,
        uid=uid,
        force_overwrite=force_overwrite,
        locspec=locspec,
        author=author,
        title=title,
        year=year,
        date=date,
    )

    logger.debug("Successfully added a new source entry")



# zettelkasten/cli.py
# type: ignore[attr-defined]

from typing import Optional

import random
from enum import Enum

import typer
from rich.console import Console

from . import __version__


class Color(str, Enum):
    white = "white"
    red = "red"
    cyan = "cyan"
    magenta = "magenta"
    yellow = "yellow"
    green = "green"


app = typer.Typer(
    name="zettelkasten",
    help="Folder based zettelkasten with a bibtex reference system and emacs "
    + "org-mode file zettels",
    add_completion=False,
)
console = Console()


def version_callback(value: bool):
    """Prints the version of the package."""
    if value:
        console.print(
            f"[yellow]zettelkasten[/] version: [bold blue]{__version__}[/]"
        )
        raise typer.Exit()


@app.callback()
def version(
    version: bool = typer.Option(
        None,
        "-v",
        "--version",
        callback=version_callback,
        # is_eager=True,
    ),
):
    """ Prints the version of the package"""


@app.command()
def add():
    """Adds a Zettel"""
    typer.echo("Adding Zettel")


@app.command()
def change():
    """Change a Zettel"""
    typer.echo("Change Zettel")



# zettelkasten/defaults.py
from typing import List, Mapping, Optional, Union

import inspect
import os
from datetime import date
from pathlib import Path

config_overwrites = [
    "config_folder",
    "config_file",
    "def_author",
    "def_title",
    "def_year",
    "def_date",
    "def_location_specifier",
    "initial_folder_structure",
    "location",
    "name_sep",
    "required_attributes",
    "sources_directory",
    # dict of key, list
    "sources_file_formats",
    # dict of key stirng and key, list
    "zettel_meta_attribute_defaults",
    # dict of key, string
    "zettel_meta_attribute_labels",
    "zettelkasten_bib_file",
]
""" Default attributes that are designed to be
:mod:`monkeypatched <zettelkasten.monkeypatch>` during zettelkasten command
executions."""

location = os.path.normpath(
    os.path.join(
        inspect.getfile(inspect.currentframe()).split(  # type: ignore
            "defaults"
        )[0],
        "..",
        "tests",
        "testkasten",
    )
)
"""
Default location of the zettelkasten. For testing purposes this states a folder
inside the installation directory. Is to be :mod:`monkeypatched
<zettelkasten.monkeypatch>` during zettelkasten command executions.
"""

def_author = "Ammon, Mathias"
"""Default author name when creating a reference entry."""

def_title = "Test Title"
""" Default source title when creating a reference entry """

def_year = date.today().year
""" Default year when creating a reference entry """

def_date = date.today().strftime("%Y-%m-%d")
""" Default date when creating a reference entry """

def_location_specifier = None
""" Default location specifier inside the source (like ``page 5``) when
creating a reference entry """


def bibliography_entry(
    source_file,
    key,
    location_specifier=def_location_specifier,
    author=def_author,
    title=def_title,
    year=def_year,
    date=def_date,
):
    """
    Default utility for creating bib entries. Located here so a dedicated user
    can overwrite this. Not designed to be monkeypatched.
    """

    entry = [
        f"@misc{{{key},\n",
        f"  author   = {{{author}}},\n",
        f"  title    = {{{title}}},\n",
        f"  year     = {{{year}}},\n",
        f"  date     = {{{date}}},\n",
        f'  url      = "file://{os.path.abspath(source_file)}",\n',
        f"  keywords = {{{location_specifier}}},\n",
        "}%",
        "\n",
    ]

    return entry


config_folder = Path.home() / ".zettelkasten.d"
"""
Path the zettelkasten configuration and rc files reside in for userfriendly
tweaking.
"""

config_file = config_folder / "zk.cfg"
"""
Path the zettelkasten configuration file reside in for userfriendly tweaking.
"""

zettel_meta_attribute_labels = {
    "uid": "#+Title:",  # required for zettel adding to work
    "category": "#+Category:",  # required for zettel adding to work
    "subcategory": "#+Subcategory:",  # required for zettel adding to work
    # optional
    "author": "#+Author:",
    "doc": "#+DOC:",
    "dole": "#+DOLE:",
    "topics": "#+Topics:",
    "tags": "#+Tags:",
}
"""
Org mode tags, used for :mod:`zettelkasten.compile` mechanics.
"""

required_attributes = (
    "uid",
    "category",
    "subcategory",
)
"""
:attr:`Zettel meta attributes <zettel_meta_attribute_labels>` nedded by
:mod:`zettelkasten.compile` to generate zettelkasten meta information.
"""

zettel_meta_attribute_defaults: Mapping[
    str, Union[Optional[str], List[Optional[str]]]
] = {
    "category": None,  # required for zettel adding to work
    "subcategory": None,  # required for zettel adding to work
    # optional
    "author": "Mathias Ammon",
    "doc": None,
    "dole": None,
    "topics": [],
    "tags": [],
}
"""
Default values for initiating the :attr:`zettel_meta_attribute_labels`
"""

name_sep = "/"
"""
Seperator between category, subcategory and zetteluid.
See :func:`zettelkasten.add.parse_zettel_name` for more details.
"""

sources_directory = "_sources"
"""
Top level folder of all reference files.
"""

zettelkasten_bib_file = "zettelkasten.bib"
"""
The zettelkasten main bibliography location. All of the references ever added
are to be found inside this file.
"""

initial_folder_structure = [
    "lobby",
    f"{sources_directory}",
    "_sources/audios",
    "_sources/images",
    "_sources/pdfs",
    "_sources/videos",
]
"""
Folder created on the initial :ref:`setup <installation_setup>`.
"""

source_file_formats = {
    "audios": [
        "mp3",
    ],
    "images": [
        "jpg",
    ],
    "pdfs": [
        "pdf",
    ],
    "videos": [
        "mp4",
    ],
}
"""
Mapping used for infering file types and respective reference file sortings
(Video files beeing copied to ``_sources/videos/`` etc.).
"""


def infer_file_type(f):
    """ Tries to enfer the file type by looking at the file ending"""
    file_ending = f.split(".")[-1]

    for folder, file_endings in source_file_formats.items():
        if file_ending in file_endings:
            return folder


state = "hardcoded"
"""
State indicating if module contents are currently monkeypatched by
:mod:`zettelkasten.monkeypatch`.
"""



# zettelkasten/setupy.py
"""
Module handling the very first use of the zettelkasten after installation.
"""

import logging
from pathlib import Path

from . import defaults

logger = logging.getLogger(__name__)


def create_config_folder(dummy_location=None):
    """
    Create the zettelkasten's tweaking folder for the configuration
    and rc files.

    Parameters
    ----------
    dummy_location: str, pathlib.Path, None, default=None
        Dummy location used for testing. Design usage is to fallback on
        :attr:`zettelkasten.defaults.config_folder`.

    Examples
    --------
    1. Manipulate the zettelkasten's defaults to not override any zk in use
       (zk does not override anything unless explicitly stated, but better
       safe than sorry in this case):

        >>> import zettelkasten.defaults as defaults
        >>> original_folder_loc = defaults.location
        >>> defaults.config_folder = 'tests/doctest_dir/.zettelkasten.d'

    2. Create the folder:

        >>> create_config_folder()

    3. Revert to default settings to not mess up any following tets

        >>> defaults.config_folder = original_folder_loc

    4. Check if is where you would expect it

        >>> from pathlib import Path
        >>> Path('tests/doctest_dir/.zettelkasten.d').is_dir()
        True
    """

    logger.debug("Initializing zettelkasten configuration file")

    if dummy_location:
        logger.debug("Config folder requested to be created at dummy location")
        logger.debug(f"{Path(dummy_location)}")

        config_folder_path = Path(dummy_location)
    else:

        logger.debug(
            "Config folder requested to be created at default location"
        )
        logger.debug(f"{defaults.config_folder}")
        config_folder_path = Path(defaults.config_folder)

    config_folder_path.mkdir(parents=True, exist_ok=True)
    logger.debug("Succesfully create zettelkasten config folder at")
    logger.debug(f"{Path(config_folder_path).resolve()}")


def create_config_file_lines():
    """ Wrapper for creating the initial config file content as lines. """
    lines = [
        "[default]\n",
        "config_folder = ~/.zettelkasten.d\n",
        "\n",
        "def_author = Ammon, Mathias\n",
        "def_title = Config Parsed Test Title\n",
        "def_location_specifier = None\n",
        "\n",
        "location = ~/zettelkasten\n",
        "\n",
        "initial_folder_structure = \n",
        "    lobby,\n",
        "    %(sources_directory)s,\n",
        "    _sources/audios,\n",
        "    _sources/images,\n",
        "    _sources/pdfs,\n",
        "    _sources/videos\n",
        "\n",
        "name_sep = /\n",
        "\n",
        "required_attributes = \n",
        "    uid,\n",
        "    category,\n",
        "    subcategory\n",
        "\n",
        "sources_directory = _sources\n",
        "\n",
        "zettelkasten_bib_file = zettelkasten.bib\n",
        "\n",
        "[source_file_formats]\n",
        "audios = \n",
        "    mp3,\n",
        "    wav\n",
        "images = \n",
        "    jpg,\n",
        "    png\n",
        "pdfs =\n",
        "    pdf,\n",
        "    odt\n",
        "videos =\n",
        "    mp4\n",
        "\n",
        "[zettel_meta_attribute_defaults]\n",
        "# required for zettel adding to work    \n",
        "category= None\n",
        "subcategory= None\n",
        "# optional\n",
        "author = Mathias Ammon\n",
        "topics =\n",
        "tags =\n",
        "doc = today\n",
        "\n",
        "[zettel_meta_attribute_labels]\n",
        "# required for zettel adding to work\n",
        "uid = #+Title:\n",
        "category = #+Category:\n",
        "subcategory = #+Subcategory:\n",
        "# optional\n",
        "author = #+Author:\n",
        "doc = #+DOC:\n",
        "dole = #+DOLE:\n",
        "topics = #+Topics:\n",
        "tags = #+Tags:\n",
    ]

    return lines


def create_config_file(dummy_location=None):
    """
    Create the zettelkasten's :ref:`cfile` for tweaking locations
    templates etc. Every attribute inside :mod:`zettelkasten.defaults` can be
    overridden/tweaked using it. The :ref:`config file's <cfile>` default
    location is ``~/.zettelkasten.d/zk.cfg``.

    Note
    ----
    The top level folder (``.zettelkasten.d/`` by default) must exist for
    this function to succeed.

    Parameters
    ----------
    dummy_location: str, pathlib.Path, None, default=None
        Dummy location used for testing. Design usage is to fallback on
        :attr:`zettelkasten.defaults.config_file`.

    Examples
    --------
    1. Manipulate the zettelkasten's defaults to not override any zk in use
       (zk does not override anything unless explicitly stated, but better
       sage than sorry in this case):

        >>> import zettelkasten.defaults as defaults
        >>> original_folder_loc = defaults.config_folder
        >>> original_file_loc = defaults.config_file
        >>> defaults.config_folder = 'tests/doctest_dir/.zettelkasten.d'
        >>> defaults.config_file = 'tests/doctest_dir/.zettelkasten.d/zk.cfg'

    2. Create the folder:

        >>> create_config_folder()
        >>> create_config_file()

    3. Revert to default settings to not mess up any following tets

        >>> defaults.config_folder = original_folder_loc
        >>> defaults.config_file = original_file_loc

    4. Check if is where you would expect it

        >>> from pathlib import Path
        >>> Path('tests/doctest_dir/.zettelkasten.d/zk.cfg').is_file()
        True
    """
    logger.debug("Initializing zettelkasten configuration file")

    if dummy_location:
        logger.debug("Config file requested to be created at dummy location")
        logger.debug(f"{Path(dummy_location)}")

        config_file_path = Path(dummy_location)
    else:

        logger.debug("Config file requested to be created at default location")
        logger.debug(f"{defaults.config_file}")
        config_file_path = Path(defaults.config_file)

    if config_file_path.is_file():
        logger.debug("Configuration file already present")
        logger.debug("Zettelkasten has been setup before.")
        logger.debug("Not overwriting the existing config.")
    else:
        with open(config_file_path, "w") as f:
            f.writelines(create_config_file_lines())

            logger.debug("Succesfully create zettelkasten config file at")
            logger.debug(f"{config_file_path}")



# zettelkasten/__main__.py


def main():
    msg = "".join(
        [
            "Folder based zettelkasten with a bibtex reference system and",
            "emacs org-mode file zettels",
        ]
    )
    print(msg)


if __name__ == "__main__":
    main()



# zettelkasten/initialize.py
"""
Module handling zettelkasten creation and recreation. Used before most of the
commands coming from the cli are executed to ensure the zettelkasten looks like
it is expected to.
"""

import logging
import os
import pathlib

from . import defaults

logger = logging.getLogger(__name__)


def structure_zettelkasten(dummy_location=None):
    """Utility to (re)structure the zettelkasten.

    :attr:`defaults.location <zettelkasten.defaults.location>` is used for
    the zettelkasten main directory and
    :attr:`defaults.initial_folder_structure
    <zettelkasten.defaults.initial_folder_structure>` for toplevel directories
    inside the zettelkasten.

    Designed to be called before executing any other command to ensure
    the zettelkasten is structured like expected by the rest of the api.

    Parameters
    ----------
    dummy_location: str, pathlib.Path, None, default=None
        Dummy location used for testing. Design usage is to fallback on
        :attr:`zettelkasten.defaults.location`.

    Examples
    --------
    Create a zettelkasten at the default location (tests directory of
    'zettelkasten'):


    >>> structure_zettelkasten()

    Check if the main directory is present:

    >>> import zettelkasten.defaults
    >>> import os
    >>> os.path.isdir(zettelkasten.defaults.location)
    True
    """
    # create zettelkasten directory
    if dummy_location:
        location = dummy_location
    else:
        location = defaults.location

    if not os.path.isdir(location):
        logger.debug(
            f"The zettelkasten directory ('{location}'"
            + "is not an existing directory"
        )
        logger.debug("... so it's created")
    else:
        logger.debug(f"Zettelkasten directory in '{location}' already exsits.")
        logger.debug("Reastablishing initial folder structure")
        logger.debug("without deleting any existing structures.")

    for directory in defaults.initial_folder_structure:
        pathlib.Path(os.path.join(location, directory)).mkdir(
            parents=True, exist_ok=True
        )

    # create the zettelkasten bib file
    zk_bib_file = os.path.join(
        location,
        defaults.sources_directory,
        defaults.zettelkasten_bib_file,
    )

    with open(zk_bib_file, "a"):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zettelkasten']

package_data = \
{'': ['*']}

install_requires = \
['rich>=9.13.0,<10.0.0', 'typer[all]>=0.3.2,<0.4.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=3.4.0,<4.0.0']}

entry_points = \
{'console_scripts': ['zettelkasten = zettelkasten.cli:app']}

setup_kwargs = {
    'name': 'zettelkasten',
    'version': '0.2.3',
    'description': 'CLI interface for managing a folder based zettelkasten using emacs org-mode file zettels and a bibtex reference system',
    'long_description': '# zettelkasten\n\n<div align="center">\n\n[![Build status](https://github.com/tZ3ma/zettelkasten/workflows/build/badge.svg?branch=master&event=push)](https://github.com/tZ3ma/zettelkasten/actions?query=workflow%3Abuild)\n[![Python Version](https://img.shields.io/pypi/pyversions/zettelkasten.svg)](https://pypi.org/project/zettelkasten/)\n[![Dependencies Status](https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg)](https://github.com/tZ3ma/zettelkasten/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/tZ3ma/zettelkasten/blob/master/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%F0%9F%9A%80-semantic%20versions-informational.svg)](https://github.com/tZ3ma/zettelkasten/releases)\n[![License](https://img.shields.io/github/license/tZ3ma/zettelkasten)](https://github.com/tZ3ma/zettelkasten/blob/master/LICENSE)\n\nFolder based zettelkasten with a bibtex reference system and emacs org-mode file zettels\n\n</div>\n\n## Very first steps\n\n### Initial\n\n1. Initialize `git` inside your repo:\n\n```bash\ngit init\n```\n\n2. If you don\'t have `Poetry` installed run:\n\n```bash\nmake download-poetry\n```\n\n3. Initialize poetry and install `pre-commit` hooks:\n\n```bash\nmake install\n```\n\n4. Upload initial code to GitHub (ensure you\'ve run `make install` to use `pre-commit`):\n\n```bash\ngit add .\ngit commit -m ":tada: Initial commit"\ngit branch -M main\ngit remote add origin https://github.com/tZ3ma/zettelkasten.git\ngit push -u origin main\n```\n\n### Initial setting up\n\n- Set up [Dependabot](https://docs.github.com/en/github/administering-a-repository/enabling-and-disabling-version-updates#enabling-github-dependabot-version-updates) to ensure you have the latest dependencies.\n- Set up [Stale bot](https://github.com/apps/stale) for automatic issue closing.\n\n### Poetry\n\nAll manipulations with dependencies are executed through Poetry. If you\'re new to it, look through [the documentation](https://python-poetry.org/docs/).\n\n<details>\n<summary>Notes about Poetry</summary>\n<p>\n\nPoetry\'s [commands](https://python-poetry.org/docs/cli/#commands) are very intuitive and easy to learn, like:\n\n- `poetry add numpy`\n- `poetry run pytest`\n- `poetry build`\n- etc\n\n</p>\n</details>\n\n### Building your package\n\nBuilding a new version of the application contains steps:\n\n- Bump the version of your package `poetry version <version>`. You can pass the new version explicitly, or a rule such as `major`, `minor`, or `patch`. For more details, refer to the [Semantic Versions](https://semver.org/) standard.\n- Make a commit to `GitHub`.\n- Create a `GitHub release`.\n- And... publish 🙂 `poetry publish --build`\n\n## What\'s next\n\nWell, that\'s up to you. I can only recommend the packages and articles that helped me.\n\nPackages:\n\n- [`Typer`](https://github.com/tiangolo/typer) is great for creating CLI applications.\n- [`Rich`](https://github.com/willmcgugan/rich) makes it easy to add beautiful formatting in the terminal.\n- [`FastAPI`](https://github.com/tiangolo/fastapi) is a type-driven asynchronous web framework.\n- [`IceCream`](https://github.com/gruns/icecream) is a little library for sweet and creamy debugging\n\nArticles:\n\n- [Open Source Guides](https://opensource.guide/)\n- [GitHub Actions Documentation](https://help.github.com/en/actions)\n- Maybe you would like to add [gitmoji](https://gitmoji.carloscuesta.me/) to commit names. This is really funny. 😄\n\n## 🚀 Features\n\nFor your development we\'ve prepared:\n\n- Supports for `Python 3.7` and higher.\n- [`Poetry`](https://python-poetry.org/) as the dependencies manager. See configuration in [`pyproject.toml`](https://github.com/tZ3ma/zettelkasten/blob/master/pyproject.toml) and [`setup.cfg`](https://github.com/tZ3ma/zettelkasten/blob/master/setup.cfg).\n- Power of [`black`](https://github.com/psf/black), [`isort`](https://github.com/timothycrosley/isort) and [`pyupgrade`](https://github.com/asottile/pyupgrade) formatters.\n- Ready-to-use [`pre-commit`](https://pre-commit.com/) hooks with formatters above.\n- Type checks with the configured [`mypy`](https://mypy.readthedocs.io).\n- Testing with [`pytest`](https://docs.pytest.org/en/latest/).\n- Docstring checks with [`darglint`](https://github.com/terrencepreilly/darglint).\n- Security checks with [`safety`](https://github.com/pyupio/safety) and [`bandit`](https://github.com/PyCQA/bandit).\n- Well-made [`.editorconfig`](https://github.com/tZ3ma/zettelkasten/blob/master/.editorconfig), [`.dockerignore`](https://github.com/tZ3ma/zettelkasten/blob/master/.dockerignore), and [`.gitignore`](https://github.com/tZ3ma/zettelkasten/blob/master/.gitignore). You don\'t have to worry about those things.\n\nFor building and deployment:\n\n- `GitHub` integration.\n- [`Makefile`](https://github.com/tZ3ma/zettelkasten/blob/master/Makefile#L89) for building routines. Everything is already set up for security checks, codestyle checks, code formatting, testing, linting, docker builds, etc. More details at [Makefile summary](#makefile-usage)).\n- [Dockerfile](https://github.com/tZ3ma/zettelkasten/blob/master/docker/Dockerfile) for your package.\n- `Github Actions` with predefined [build workflow](https://github.com/tZ3ma/zettelkasten/blob/master/.github/workflows/build.yml) as the default CI/CD.\n- Always up-to-date dependencies with [`@dependabot`](https://dependabot.com/) (You will only [enable it](https://docs.github.com/en/github/administering-a-repository/enabling-and-disabling-version-updates#enabling-github-dependabot-version-updates)).\n- Automatic drafts of new releases with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter). It creates a list of changes based on labels in merged `Pull Requests`. You can see labels (aka `categories`) in [`release-drafter.yml`](https://github.com/tZ3ma/zettelkasten/blob/master/.github/release-drafter.yml). Works perfectly with [Semantic Versions](https://semver.org/) specification.\n\nFor creating your open source community:\n\n- Ready-to-use [Pull Requests templates](https://github.com/tZ3ma/zettelkasten/blob/master/.github/PULL_REQUEST_TEMPLATE.md) and several [Issue templates](https://github.com/tZ3ma/zettelkasten/tree/master/.github/ISSUE_TEMPLATE).\n- Files such as: `LICENSE`, `CONTRIBUTING.md`, `CODE_OF_CONDUCT.md`, and `SECURITY.md` are generated automatically.\n- [`Stale bot`](https://github.com/apps/stale) that closes abandoned issues after a period of inactivity. (You will only [need to setup free plan](https://github.com/marketplace/stale)). Configuration is [here](https://github.com/tZ3ma/zettelkasten/blob/master/.github/.stale.yml).\n- [Semantic Versions](https://semver.org/) specification with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter).\n\n## Installation\n\n```bash\npip install -U zettelkasten\n```\n\nor install with `Poetry`\n\n```bash\npoetry add zettelkasten\n```\n\nThen you can run\n\n```bash\nzettelkasten --help\n```\n\n```bash\nzettelkasten --name Roman\n```\n\nor if installed with `Poetry`:\n\n```bash\npoetry run zettelkasten --help\n```\n\n```bash\npoetry run zettelkasten --name Roman\n```\n\n### Makefile usage\n\n[`Makefile`](https://github.com/tZ3ma/zettelkasten/blob/master/Makefile) contains many functions for fast assembling and convenient work.\n\n<details>\n<summary>1. Download Poetry</summary>\n<p>\n\n```bash\nmake download-poetry\n```\n\n</p>\n</details>\n\n<details>\n<summary>2. Install all dependencies and pre-commit hooks</summary>\n<p>\n\n```bash\nmake install\n```\n\nIf you do not want to install pre-commit hooks, run the command with the NO_PRE_COMMIT flag:\n\n```bash\nmake install NO_PRE_COMMIT=1\n```\n\n</p>\n</details>\n\n<details>\n<summary>3. Check the security of your code</summary>\n<p>\n\n```bash\nmake check-safety\n```\n\nThis command launches a `Poetry` and `Pip` integrity check as well as identifies security issues with `Safety` and `Bandit`. By default, the build will not crash if any of the items fail. But you can set `STRICT=1` for the entire build, or you can configure strictness for each item separately.\n\n```bash\nmake check-safety STRICT=1\n```\n\nor only for `safety`:\n\n```bash\nmake check-safety SAFETY_STRICT=1\n```\n\nmultiple\n\n```bash\nmake check-safety PIP_STRICT=1 SAFETY_STRICT=1\n```\n\n> List of flags for `check-safety` (can be set to `1` or `0`): `STRICT`, `POETRY_STRICT`, `PIP_STRICT`, `SAFETY_STRICT`, `BANDIT_STRICT`.\n\n</p>\n</details>\n\n<details>\n<summary>4. Check the codestyle</summary>\n<p>\n\nThe command is similar to `check-safety` but to check the code style, obviously. It uses `Black`, `Darglint`, `Isort`, and `Mypy` inside.\n\n```bash\nmake check-style\n```\n\nIt may also contain the `STRICT` flag.\n\n```bash\nmake check-style STRICT=1\n```\n\n> List of flags for `check-style` (can be set to `1` or `0`): `STRICT`, `BLACK_STRICT`, `DARGLINT_STRICT`, `ISORT_STRICT`, `MYPY_STRICT`.\n\n</p>\n</details>\n\n<details>\n<summary>5. Run all the codestyle formaters</summary>\n<p>\n\nCodestyle uses `pre-commit` hooks, so ensure you\'ve run `make install` before.\n\n```bash\nmake codestyle\n```\n\n</p>\n</details>\n\n<details>\n<summary>6. Run tests</summary>\n<p>\n\n```bash\nmake test\n```\n\n</p>\n</details>\n\n<details>\n<summary>7. Run all the linters</summary>\n<p>\n\n```bash\nmake lint\n```\n\nthe same as:\n\n```bash\nmake test && make check-safety && make check-style\n```\n\n> List of flags for `lint` (can be set to `1` or `0`): `STRICT`, `POETRY_STRICT`, `PIP_STRICT`, `SAFETY_STRICT`, `BANDIT_STRICT`, `BLACK_STRICT`, `DARGLINT_STRICT`, `ISORT_STRICT`, `MYPY_STRICT`.\n\n</p>\n</details>\n\n<details>\n<summary>8. Build docker</summary>\n<p>\n\n```bash\nmake docker\n```\n\nwhich is equivalent to:\n\n```bash\nmake docker VERSION=latest\n```\n\nMore information [here](https://github.com/tZ3ma/zettelkasten/tree/master/docker).\n\n</p>\n</details>\n\n<details>\n<summary>9. Cleanup docker</summary>\n<p>\n\n```bash\nmake clean_docker\n```\n\nor to remove all build\n\n```bash\nmake clean\n```\n\nMore information [here](https://github.com/tZ3ma/zettelkasten/tree/master/docker).\n\n</p>\n</details>\n\n## 📈 Releases\n\nYou can see the list of available releases on the [GitHub Releases](https://github.com/tZ3ma/zettelkasten/releases) page.\n\nWe follow [Semantic Versions](https://semver.org/) specification.\n\nWe use [`Release Drafter`](https://github.com/marketplace/actions/release-drafter). As pull requests are merged, a draft release is kept up-to-date listing the changes, ready to publish when you’re ready. With the categories option, you can categorize pull requests in release notes using labels.\n\nFor Pull Request this labels are configured, by default:\n\n|               **Label**               |  **Title in Releases**  |\n| :-----------------------------------: | :---------------------: |\n|       `enhancement`, `feature`        |       🚀 Features       |\n| `bug`, `refactoring`, `bugfix`, `fix` | 🔧 Fixes & Refactoring  |\n|       `build`, `ci`, `testing`        | 📦 Build System & CI/CD |\n|              `breaking`               |   💥 Breaking Changes   |\n|            `documentation`            |    📝 Documentation     |\n|            `dependencies`             | ⬆️ Dependencies updates |\n\nYou can update it in [`release-drafter.yml`](https://github.com/tZ3ma/zettelkasten/blob/master/.github/release-drafter.yml).\n\nGitHub creates the `bug`, `enhancement`, and `documentation` labels for you. Dependabot creates the `dependencies` label. Create the remaining labels on the Issues tab of your GitHub repository, when you need them.\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/tZ3ma/zettelkasten)](https://github.com/tZ3ma/zettelkasten/blob/master/LICENSE)\n\nThis project is licensed under the terms of the `MIT` license. See [LICENSE](https://github.com/tZ3ma/zettelkasten/blob/master/LICENSE) for more details.\n\n## 📃 Citation\n\n```\n@misc{zettelkasten,\n  author = {tZ3ma},\n  title = {Folder based zettelkasten with a bibtex reference system and emacs org-mode file zettels},\n  year = {2021},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/tZ3ma/zettelkasten}}\n}\n```\n\n## Credits\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template).\n',
    'author': 'tZ3ma',
    'author_email': 'mathias.ammon@tuhh.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tZ3ma/zettelkasten',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
